import 'dart:core';

/// Error thrown when the return value of a function is unacceptable.
class ResultError extends Error {
  /// Whether value was provided.
  final bool _hasValue;

  /// The invalid value.
  final dynamic invalidValue;

  /// Name of the invalid result, if available.
  final String? name;

  /// Message describing the problem.
  final dynamic message;

  /// Creates an error with [message] describing the problem
  /// with a return value.
  ///
  /// Existing code may be using `message` to hold the invalid value.
  /// If the `message` is not a [String], it is assumed to be a
  /// value instead of a message.
  @pragma("vm:entry-point")
  ResultError([this.message])
      : _hasValue = false,
        invalidValue = null,
        name = null;

  /// Creates an result error for a `null` result that must not be `null`.
  ResultError.notNull([this.name])
      : _hasValue = false,
        message = "Must not be null",
        invalidValue = null;

  /// Throws if [result] is `null`.
  ///
  /// If [name] is supplied, it is used as the parameter name
  /// in the error message.
  ///
  /// Returns the [result] if it is not null.
  static T checkNotNull<T>(T? result, [String? name]) {
    if (null == result) throw ResultError.notNull(name);
    return result;
  }

  // Helper functions for toString overridden in subclasses.
  String get _errorName => "Invalid result${!_hasValue ? "(s)" : ""}";
  String get _errorExplanation => "";

  @override
  String toString() {
    final name = this.name;
    final nameString = (null == name) ? "" : " ($name)";

    final message = this.message as Object?;
    final messageString = (null == message) ? "" : ": $message";

    final prefix = "$_errorName$nameString$messageString";

    if (!_hasValue) return prefix;

    // If we know the invalid value, we can try to describe the problem.
    final explanation = _errorExplanation;
    final errorValue = Error.safeToString(invalidValue);

    return "$prefix$explanation: $errorValue";
  }
}
