import 'dart:async';
import 'dart:core';

import 'package:flutter/foundation.dart';
import 'package:flutter/services.dart';

import 'exceptions.dart';
import 'javascript.dart';
import 'typedefs.dart';
import 'web_resource_error.dart';

const String _namespace =
    "com.twelve_ampere.webview_without_view.WebviewWithoutViewPlugin";
const String _methodChannelId = "$_namespace.methods";

MethodChannel? _channel;
_Settings? _settings;

_Settings get settings => _settings ??= _Settings._();

/// Before using any functions, this plugin has to be
/// initialized.
/// Nothing happens if [initialize] is called more than
/// once.
Future<bool> initialize([String? initialUrl]) async {
  if (null != _channel) return true;

  _channel = const MethodChannel(_methodChannelId);

  final params = {
    "usesHybridComposition": true,
    "settings": settings._toMap(),
    if (null != initialUrl) "initialUrl": initialUrl
  };

  if (!(await _channel!.invokeMethod<bool>(
        "initialize",
        params,
      ) ??
      false)) return false;

  _channel!.setMethodCallHandler(_onMethodCallHandler);

  // ignore: invalid_use_of_protected_member
  if (!settings._notifier.hasListeners)
    settings._notifier.addListener(_onSettingsNotify);

  return true;
}

/// Calling [dispose] cuts the connection to the
/// platform code and therefore removes the functionality
/// of any functions in this library.
Future<void> dispose() async {
  settings._notifier.removeListener(_onSettingsNotify);
  await _channel?.invokeMethod("dispose");
  _channel?.setMethodCallHandler(null);
  _channel = null;
}

/// Loads the given URL with additional HTTP headers, specified
/// as a map from name to value.
Future<void> loadUrl(
  String url, [
  Map<String, String>? additionalHeaders,
]) async =>
    _channel?.invokeMethod("loadUrl", {
      "url": url,
      "additionalHeaders": additionalHeaders,
    });

/// Asynchronously evaluates JavaScript in the context of
/// the currently loaded page.
///
/// Throws a [ResultError], if the result of the evaluated
/// javascript is `null`.
Future<String> evaluateJavascript(String script) async {
  final result = await _channel?.invokeMethod<String>("evaluateJavascript", {
    "script": script,
  });

  return ResultError.checkNotNull<String>(result);
}

/// Removes all cookies.
Future<void> clearCookies() async => _channel?.invokeMethod("clearCookies");

/// Clears the resource cache and all storage currently being
/// used by the JavaScript storage APIs. This includes the
/// Application Cache, Web SQL Database and the HTML5 Web Storage APIs.
Future<void> clearCache() async => _channel?.invokeMethod("clearCache");

/// Reloads the current URL.
Future<void> reload() async => _channel?.invokeMethod("reload");

Future<void> _updateSettings() async => _channel?.invokeMethod(
      "_updateSettings",
      settings._toMap(),
    );

Future<dynamic> _onMethodCallHandler(MethodCall call) async {
  if ("javascriptChannelMessage" == call.method) {
    final channelName = call.arguments["channel"] as String;
    final message = call.arguments["message"] as String;
    settings.javascriptChannel?.onMessageReceived(JavascriptMessage(message));
    return null;
  }

  if ("onProgress" == call.method) {
    final progress = call.arguments["progress"] as int;
    settings.onProgress?.call(progress);
    return null;
  }

  if ("onPageStarted" == call.method) {
    final url = call.arguments["url"] as String;
    settings.onPageStarted?.call(url);
    return null;
  }

  if ("onPageFinished" == call.method) {
    final url = call.arguments["url"] as String;
    settings.onPageFinished?.call(url);
    return null;
  }

  if ("onWebResourceError" == call.method) {
    final errorCode = call.arguments["errorCode"] as int;
    final description = call.arguments["description"] as String;
    final failingUrl = call.arguments["failingUrl"] as String;
    final errorTypeString = call.arguments["errorType"] as String?;
    final domain = call.arguments["domain"] as String?;

    settings.onWebResourceError?.call(WebResourceError(
      errorCode: errorCode,
      description: description,
      failingUrl: failingUrl,
      domain: domain,
      errorType: null == errorTypeString
          ? null
          : WebResourceErrorType.values.firstWhere(
              (element) =>
                  "$WebResourceErrorType.$errorTypeString" ==
                  element.toString(),
            ),
    ));
    return null;
  }

  if ("navigationRequest" == call.method) {
    final url = call.arguments["url"] as String;
    final isForMainFrame = call.arguments["isForMainFrame"] as bool;

    if (NavigationDecision.navigate ==
        await settings.navigationDelegate!(NavigationRequest._(
          url: url,
          isForMainFrame: isForMainFrame,
        ))) return true;

    return false;
  }

  throw MissingPluginException("${call.method} was invoked but has no handler");
}

void _onSettingsNotify() => _updateSettings();

class _Settings {
  final ChangeNotifier _notifier;

  String? _initialUrl;
  JavascriptMode _javascriptMode;
  JavascriptChannel? _javascriptChannel;
  bool _allowsInlineMediaPlayback;
  bool _debuggingEnabled;
  String? _userAgent;
  AutoMediaPlaybackPolicy _initialMediaPlaybackPolicy;

  /// A delegate function that decides how to handle
  /// navigation actions. When a navigation is initiated by the WebView
  /// (e.g when a user clicks a link) this delegate is called and has to decide
  /// how to proceed with the navigation. See [NavigationDecision] for possible
  /// decisions the delegate can take. When null all navigation actions are allowed.
  ///
  /// * *Caveats on Android*:
  ///   * Navigation actions targeted to the main frame can be intercepted,
  ///     navigation actions targeted to subframes are allowed regardless of
  ///     the value returned by this delegate.
  ///
  ///   * Setting a navigationDelegate makes the WebView treat all navigations
  ///     as if they were triggered by a user gesture, this disables some of
  ///     Chromium's security mechanisms. A navigationDelegate should only be
  ///     set when loading trusted content.
  ///
  ///   * On Android WebView versions earlier than 67(most devices running at least
  ///     Android L+ should have a later version):
  ///
  ///     * When a navigationDelegate is set pages with frames are not properly
  ///       handled by the webview, and frames will be opened in the main frame.
  ///
  ///     * When a navigationDelegate is set HTTP requests do not include the
  ///       HTTP referer header.
  NavigationDelegate? navigationDelegate;

  /// Invoked when a page starts loading.
  PageStartedCallback? onPageStarted;

  /// Invoked when a page has finished loading.
  /// This is invoked only for the main frame.
  /// When [onPageFinished] is invoked on Android, the page being rendered may
  /// not be updated yet.
  /// When invoked on iOS or Android, any Javascript code that is embedded
  /// directly in the HTML has been loaded and code injected with
  /// [evaluateJavascript] can assume this.
  PageFinishedCallback? onPageFinished;

  /// Invoked when a page is loading.
  PageLoadingCallback? onProgress;

  /// Invoked when a web resource has failed to load.
  /// This can be called for any resource (iframe, image, etc.), not just for
  /// the main page.
  WebResourceErrorCallback? onWebResourceError;

  _Settings._()
      : _javascriptMode = JavascriptMode.disabled,
        _allowsInlineMediaPlayback = true,
        _debuggingEnabled = false,
        _initialMediaPlaybackPolicy =
            AutoMediaPlaybackPolicy.requireUserActionForAllMediaTypes,
        _notifier = ChangeNotifier();

  Map<String, Object?> _toMap() => {
        "jsMode": javascriptMode.index,
        "debuggingEnabled": debuggingEnabled,
        "hasProgressTracking": hasProgressTracking,
        "hasNavigationDelegate": hasNavigationDelegate,
        "allowsInlineMediaPlayback": allowsInlineMediaPlayback,
        "autoMediaPlaybackPolicy": settings.autoMediaPlaybackPolicy.index,
        "userAgent": userAgent,
        if (null != javascriptChannel) "jsChannelName": javascriptChannel!.name,
      };

  bool get hasProgressTracking => null != settings.onProgress;

  bool get hasNavigationDelegate => null != settings.navigationDelegate;

  /// The value used for the HTTP User-Agent: request header.
  /// When null the platform's webview default is used for the User-Agent header.
  /// When the [WebView] is rebuilt with a different `userAgent`, the page
  /// reloads and the request uses the new User Agent.
  /// When [WebViewController.goBack] is called after changing `userAgent`
  /// the previous `userAgent` value is used until the page is reloaded.
  /// This field is ignored on iOS versions prior to 9 as the platform does
  /// not support a custom user agent.
  String? get userAgent => _userAgent;

  /// The initial URL to load.
  String? get initialUrl => _initialUrl;

  /// Whether Javascript execution is enabled.
  /// Defaults to [JavascriptMode.disabled]
  JavascriptMode get javascriptMode => _javascriptMode;

  /// The set of [JavascriptChannel]s available to
  /// JavaScript code running in the web view.
  /// For each [JavascriptChannel] in the set, a channel object is made
  /// available for the JavaScript code in a window property named
  /// [JavascriptChannel.name].
  /// The JavaScript code can then call `postMessage` on that object to
  /// send a message that will be passed to [JavascriptChannel.onMessageReceived].
  /// For example for the following JavascriptChannel:
  /// ```dart
  /// JavascriptChannel(name: 'Print', onMessageReceived: (JavascriptMessage message) { print(message.message); });
  /// ```
  /// JavaScript code can call:
  /// ```javascript
  /// Print.postMessage('Hello');
  /// ```
  /// To asynchronously invoke the message handler which will print the
  /// message to standard output.
  /// Adding a new JavaScript channel only takes affect after the next page is loaded.
  /// Set values must not be null. A [JavascriptChannel.name] cannot be the same
  /// for multiple channels in the list.
  /// A null value is equivalent to an empty set.
  JavascriptChannel? get javascriptChannel => _javascriptChannel;

  /// Controls whether inline playback of HTML5 videos is allowed
  /// on iOS. This field is ignored on Android because Android
  /// allows it by default.
  ///
  /// Defaults to `true`.
  bool get allowsInlineMediaPlayback => _allowsInlineMediaPlayback;

  /// Controls whether WebView debugging is enabled.
  /// Setting this to true enables [WebView debugging on Android](https://developers.google.com/web/tools/chrome-devtools/remote-debugging/).
  /// WebView debugging is enabled by default in dev builds on iOS.
  /// Defaults to `false`.
  ///
  /// * *To debug WebViews on iOS*:
  ///   * Enable developer options (Open Safari, go to Preferences ->
  ///     Advanced and make sure "Show Develop Menu in Menubar" is on.)
  ///
  ///   * From the Menu-bar (of Safari) select Develop -> iPhone Simulator
  ///     -> <your webview page>
  bool get debuggingEnabled => _debuggingEnabled;

  /// Which restrictions apply on automatic media playback.
  /// This initial value is applied to the platform's webview upon creation.
  /// Any following changes to this parameter are ignored (as long as the
  /// state of the [WebView] is preserved).
  ///
  /// Defaults to [AutoMediaPlaybackPolicy.requireUserActionForAllMediaTypes].
  AutoMediaPlaybackPolicy get autoMediaPlaybackPolicy =>
      _initialMediaPlaybackPolicy;

  set userAgent(String? agent) {
    _userAgent = agent;
    _notify();
  }

  set javascriptMode(JavascriptMode mode) {
    _javascriptMode = mode;
    _notify();
  }

  set javascriptChannel(JavascriptChannel? channel) {
    _javascriptChannel = channel;
    _notify();
  }

  set allowsInlineMediaPlayback(bool allowed) {
    _allowsInlineMediaPlayback = allowed;
    _notify();
  }

  set debuggingEnabled(bool enabled) {
    _debuggingEnabled = enabled;
    _notify();
  }

  set autoMediaPlaybackPolicy(AutoMediaPlaybackPolicy policy) {
    _initialMediaPlaybackPolicy = policy;
    _notify();
  }

  void _notify() {
    // ignore: invalid_use_of_protected_member, invalid_use_of_visible_for_testing_member
    _notifier.notifyListeners();
  }

  /// If any parameters differ from their current value,
  /// then they are updated and the all listeners are
  /// notified.
  void update({
    String? initialUrl,
    JavascriptMode javascriptMode = JavascriptMode.disabled,
    JavascriptChannel? javascriptChannel,
    bool allowsInlineMediaPlayback = false,
    bool debuggingEnabled = false,
    String? userAgent,
    AutoMediaPlaybackPolicy initialMediaPlaybackPolicy =
        AutoMediaPlaybackPolicy.requireUserActionForAllMediaTypes,
  }) {
    bool notify = false;

    if (initialUrl != _initialUrl) {
      _initialUrl = initialUrl;
      notify = true;
    }

    if (javascriptMode != _javascriptMode) {
      _javascriptMode = javascriptMode;
      notify = true;
    }

    if (javascriptChannel != _javascriptChannel) {
      _javascriptChannel = javascriptChannel;
      notify = true;
    }

    if (allowsInlineMediaPlayback != _allowsInlineMediaPlayback) {
      _allowsInlineMediaPlayback = allowsInlineMediaPlayback;
      notify = true;
    }

    if (debuggingEnabled != _debuggingEnabled) {
      _debuggingEnabled = debuggingEnabled;
      notify = true;
    }

    if (userAgent != _userAgent) {
      _userAgent = userAgent;
      notify = true;
    }

    if (initialMediaPlaybackPolicy != _initialMediaPlaybackPolicy) {
      _initialMediaPlaybackPolicy = initialMediaPlaybackPolicy;
      notify = true;
    }

    if (notify) _notify();
  }
}

/// Information about a navigation action that is about to be executed.
class NavigationRequest {
  NavigationRequest._({
    required this.url,
    required this.isForMainFrame,
  });

  /// The URL that will be loaded if the navigation is executed.
  final String url;

  /// Whether the navigation request is to be loaded as the main frame.
  final bool isForMainFrame;

  @override
  String toString() {
    return '$runtimeType(url: $url, isForMainFrame: $isForMainFrame)';
  }
}

/// A decision on how to handle a navigation request.
enum NavigationDecision {
  /// Prevent the navigation from taking place.
  prevent,

  /// Allow the navigation to take place.
  navigate,
}

/// Specifies possible restrictions on automatic media playback.
///
/// This is typically used in [WebView.autoMediaPlaybackPolicy].
// The method channel implementation is marshalling this enum to the
// value's index, so the order  is important.
enum AutoMediaPlaybackPolicy {
  /// Starting any kind of media playback requires a user action.
  ///
  /// For example: JavaScript code cannot start playing media unless the code was executed
  /// as a result of a user action (like a touch event).
  requireUserActionForAllMediaTypes,

  /// Starting any kind of media playback is always allowed.
  ///
  /// For example: JavaScript code that's triggered when the page is loaded can start playing
  /// video or audio.
  alwaysAllow,
}
